@@ -17,6 +17,8 @@ import uuid
 from collections import namedtuple
 from datetime import date
 from datetime import datetime
+from datetime import timedelta
+from datetime import timezone
 
 import pytest
 import requests
@@ -138,6 +140,16 @@ uids_used = (
     "20010712T182145Z-123401@example.com",
     "20070313T123432Z-456553@example.com",
     "20080712T182145Z-123401@example.com",
+    "19970901T130000Z-123403@example.com",
+    "20010712T182145Z-123401@example.com",
+    "20080712T182145Z-123401@example.com",
+    "takeoutthethrash",
+    "ctuid1",
+    "ctuid2",
+    "ctuid3",
+    "ctuid4",
+    "ctuid5",
+    "ctuid6",
 )
 ## TODO: todo7 is an item without uid.  Should be taken care of somehow.
 
@@ -263,6 +275,7 @@ BEGIN:VTODO
 DTSTAMP:19980101T130000Z
 DTSTART:19980415T133000Z
 DUE:19980516T045959Z
+STATUS:NEEDS-ACTION
 SUMMARY:Get stuck with Netfix and forget about the tax income declaration
 CLASS:CONFIDENTIAL
 CATEGORIES:FAMILY
@@ -278,6 +291,7 @@ BEGIN:VTODO
 UID:takeoutthethrash
 DTSTAMP:20221013T151313Z
 DTSTART:20221017T065500Z
+STATUS:NEEDS-ACTION
 DURATION:PT10M
 SUMMARY:Take out the thrash before the collectors come.
 RRULE:FREQ=WEEKLY;BYDAY=MO;BYHOUR=6;BYMINUTE=55;COUNT=3
@@ -478,6 +492,8 @@ class RepeatedFunctionalTestsBaseClass(object):
     def setup_method(self):
         logging.debug("############## test setup")
         self.incompatibilities = set()
+        self.cleanup_regime = self.server_params.get("cleanup", "light")
+        self.calendars_used = []
 
         for flag in self.server_params.get("incompatibilities", []):
             assert flag in compatibility_issues.incompatibility_description
@@ -499,13 +515,7 @@ class RepeatedFunctionalTestsBaseClass(object):
         else:
             self.principal = self.caldav.principal()
 
-        if not self.check_compatibility_flag("read_only"):
-            logging.debug(
-                "## going to tear down old test calendars, "
-                "in case teardown_method wasn't properly executed "
-                "last time tests were run"
-            )
-            self._teardown_method()
+        self._cleanup("pre")
 
         if self.check_compatibility_flag("object_by_uid_is_broken"):
             import caldav.objects
@@ -520,13 +530,18 @@ class RepeatedFunctionalTestsBaseClass(object):
         logging.debug("############################")
         logging.debug("############## test teardown_method")
         logging.debug("############################")
-        self._teardown_method()
+        self._cleanup("post")
         logging.debug("############## test teardown_method done")
 
-    def _teardown_method(self):
+    def _cleanup(self, mode=None):
+        if self.cleanup_regime in ("pre", "post") and self.cleanup_regime != mode:
+            return
         if self.check_compatibility_flag("read_only"):
             return  ## no cleanup needed
-        if self.check_compatibility_flag("no_mkcalendar"):
+        if (
+            self.check_compatibility_flag("no_mkcalendar")
+            or self.cleanup_regime == "thorough"
+        ):
             for uid in uids_used:
                 try:
                     obj = self._fixCalendar().object_by_uid(uid)
@@ -538,21 +553,30 @@ class RepeatedFunctionalTestsBaseClass(object):
                         "Something went kaboom while deleting event", exc_info=True
                     )
             return
-        for name in ("Yep", "Yapp", "Yølp", self.testcal_id, self.testcal_id2):
-            try:
-                cal = self.principal.calendar(name=name)
-            except:
-                cal = self.principal.calendar(cal_id=name)
-            try:
-                if self.check_compatibility_flag("sticky_events"):
-                    try:
-                        for goo in cal.objects():
-                            goo.delete()
-                    except:
-                        pass
-                cal.delete()
-            except:
-                pass
+        for cal in self.calendars_used:
+            cal.delete()
+        if self.check_compatibility_flag("unique_calendar_ids") and mode == "pre":
+            a = self._teardownCalendar(name="Yep")
+        if mode == "post":
+            for calid in (self.testcal_id, self.testcal_id2):
+                self._teardownCalendar(cal_id=calid)
+        if self.cleanup_regime == "thorough":
+            for name in ("Yep", "Yapp", "Yølp", self.testcal_id, self.testcal_id2):
+                self._teardownCalendar(name=name)
+                self._teardownCalendar(cal_id=name)
+
+    def _teardownCalendar(self, name=None, cal_id=None):
+        try:
+            cal = self.principal.calendar(name=name, cal_id=cal_id)
+            if self.check_compatibility_flag("sticky_events"):
+                try:
+                    for goo in cal.objects():
+                        goo.delete()
+                except:
+                    pass
+            cal.delete()
+        except:
+            pass
 
     def _fixCalendar(self, **kwargs):
         """
@@ -579,9 +603,22 @@ class RepeatedFunctionalTestsBaseClass(object):
                 self._default_calendar = calendars[0]
             return self._default_calendar
         else:
-            return self.principal.make_calendar(
-                name="Yep", cal_id=self.testcal_id, **kwargs
+            if not self.check_compatibility_flag(
+                "unique_calendar_ids"
+            ) and self.cleanup_regime in ("light", "pre"):
+                self._teardownCalendar(cal_id=self.testcal_id)
+            if self.check_compatibility_flag("no_displayname"):
+                name = None
+            else:
+                name = "Yep"
+            ret = self.principal.make_calendar(
+                name=name, cal_id=self.testcal_id, **kwargs
             )
+            ## TEMP - checking that the calendar works
+            ret.events()
+            if self.cleanup_regime == "post":
+                self.calendars_used.append(ret)
+            return ret
 
     def testSupport(self):
         """
@@ -661,14 +698,18 @@ class RepeatedFunctionalTestsBaseClass(object):
         assert c.url is not None
         assert len(self.principal.calendars()) != 0
 
+        str_ = str(c)
+        repr_ = repr(c)
+
         ## Not sure if those asserts make much sense, the main point here is to exercise
         ## the __str__ and __repr__ methods on the Calendar object.
-        name = c.get_property(dav.DisplayName(), use_cached=True)
-        if not name:
-            name = c.url
-        assert str(name) == str(c)
-        assert str(name) in repr(c)
+        if not self.check_compatibility_flag("no_displayname"):
+            name = c.get_property(dav.DisplayName(), use_cached=True)
+            if not name:
+                name = c.url
+            assert str(name) == str_
         assert "Calendar" in repr(c)
+        assert str(c.url) in repr(c)
 
     def testProxy(self):
         if self.caldav.url.scheme == "https":
@@ -735,6 +776,10 @@ class RepeatedFunctionalTestsBaseClass(object):
     def testCreateDeleteCalendar(self):
         self.skip_on_compatibility_flag("no_mkcalendar")
         self.skip_on_compatibility_flag("read_only")
+        if not self.check_compatibility_flag(
+            "unique_calendar_ids"
+        ) and self.cleanup_regime in ("light", "pre"):
+            self._teardownCalendar(cal_id=self.testcal_id)
         c = self.principal.make_calendar(name="Yep", cal_id=self.testcal_id)
         assert c.url is not None
         events = c.events()
@@ -776,6 +821,8 @@ class RepeatedFunctionalTestsBaseClass(object):
         ) and not self.check_compatibility_flag("no_displayname"):
             # We should be able to access the calender through the name
             c2 = self.principal.calendar(name="Yep")
+            ## may break if we have multiple calendars with the same name
+            assert c2.url == c.url
             events2 = c2.events()
             assert len(events2) == 1
             assert events2[0].url == events[0].url
@@ -785,9 +832,11 @@ class RepeatedFunctionalTestsBaseClass(object):
             dtstart=datetime(2015, 10, 10, 8, 7, 6),
             summary="This is a test event",
             dtend=datetime(2016, 10, 10, 9, 8, 7),
+            uid="ctuid1",
         )
         events = c.events()
         assert len(events) == len(existing_events) + 2
+        ev2.delete()
 
     def testCalendarByFullURL(self):
         """
@@ -880,6 +929,7 @@ class RepeatedFunctionalTestsBaseClass(object):
         my_changed_objects = c.objects_by_sync_token(
             sync_token=my_changed_objects.sync_token
         )
+
         if not self.check_compatibility_flag("fragile_sync_tokens"):
             assert len(list(my_changed_objects)) == 0
 
@@ -1036,6 +1086,11 @@ class RepeatedFunctionalTestsBaseClass(object):
     def testLoadEvent(self):
         self.skip_on_compatibility_flag("read_only")
         self.skip_on_compatibility_flag("no_mkcalendar")
+        if not self.check_compatibility_flag(
+            "unique_calendar_ids"
+        ) and self.cleanup_regime in ("light", "pre"):
+            self._teardownCalendar(cal_id=self.testcal_id)
+            self._teardownCalendar(cal_id=self.testcal_id2)
         c1 = self.principal.make_calendar(name="Yep", cal_id=self.testcal_id)
         c2 = self.principal.make_calendar(name="Yapp", cal_id=self.testcal_id2)
         e1_ = c1.save_event(ev1)
@@ -1045,10 +1100,22 @@ class RepeatedFunctionalTestsBaseClass(object):
         assert e1.url == e1_.url
         if not self.check_compatibility_flag("event_by_url_is_broken"):
             e1.load()
+        if (
+            not self.check_compatibility_flag("unique_calendar_ids")
+            and self.cleanup_regime == "post"
+        ):
+            self._teardownCalendar(cal_id=self.testcal_id)
+            self._teardownCalendar(cal_id=self.testcal_id2)
 
     def testCopyEvent(self):
         self.skip_on_compatibility_flag("read_only")
         self.skip_on_compatibility_flag("no_mkcalendar")
+        if not self.check_compatibility_flag(
+            "unique_calendar_ids"
+        ) and self.cleanup_regime in ("light", "pre"):
+            self._teardownCalendar(cal_id=self.testcal_id)
+            self._teardownCalendar(cal_id=self.testcal_id2)
+
         ## Let's create two calendars, and populate one event on the first calendar
         c1 = self.principal.make_calendar(name="Yep", cal_id=self.testcal_id)
         c2 = self.principal.make_calendar(name="Yapp", cal_id=self.testcal_id2)
@@ -1092,6 +1159,13 @@ class RepeatedFunctionalTestsBaseClass(object):
         else:
             assert len(c1.events()) == 2
 
+        if (
+            not self.check_compatibility_flag("unique_calendar_ids")
+            and self.cleanup_regime == "post"
+        ):
+            self._teardownCalendar(cal_id=self.testcal_id)
+            self._teardownCalendar(cal_id=self.testcal_id2)
+
     def testCreateCalendarAndEventFromVobject(self):
         self.skip_on_compatibility_flag("read_only")
         c = self._fixCalendar()
@@ -1128,7 +1202,10 @@ class RepeatedFunctionalTestsBaseClass(object):
 
         ## Search without any parameters should yield everything on calendar
         all_events = c.search()
-        assert len(all_events) == 3
+        if self.check_compatibility_flag("search_needs_comptype"):
+            assert len(all_events) <= 3
+        else:
+            assert len(all_events) == 3
 
         ## Search with comp_class set to Event should yield all events on calendar
         all_events = c.search(comp_class=Event)
@@ -1161,18 +1238,35 @@ class RepeatedFunctionalTestsBaseClass(object):
         if not self.check_compatibility_flag("text_search_not_working"):
             assert len(some_events) == 1
 
+        ## not defined
+        some_events = c.search(comp_class=Event, no_class=True)
+        ## ev1, ev3 should be returned
+        ## or perhaps not,
+        ## ref https://gitlab.com/davical-project/davical/-/issues/281#note_1265743591
+        ## PUBLIC is default, so maybe no events should be returned?
+        if not self.check_compatibility_flag("isnotdefined_not_working"):
+            assert len(some_events) == 2
+
+        some_events = c.search(comp_class=Event, no_category=True)
+        ## ev1, ev3 should be returned
+        if not self.check_compatibility_flag("isnotdefined_not_working"):
+            assert len(some_events) == 2
+
+        some_events = c.search(comp_class=Event, no_dtend=True)
+        ## evr should be returned
+        if not self.check_compatibility_flag("isnotdefined_not_working"):
+            assert len(some_events) == 1
+
+        self.skip_on_compatibility_flag("text_search_not_working")
+
         ## category
         if not self.check_compatibility_flag("radicale_breaks_on_category_search"):
 
             some_events = c.search(comp_class=Event, category="PERSONAL")
-            if not self.check_compatibility_flag(
-                "category_search_yields_nothing"
-            ) and not self.check_compatibility_flag("text_search_not_working"):
+            if not self.check_compatibility_flag("category_search_yields_nothing"):
                 assert len(some_events) == 1
             some_events = c.search(comp_class=Event, category="personal")
-            if not self.check_compatibility_flag(
-                "category_search_yields_nothing"
-            ) and not self.check_compatibility_flag("text_search_not_working"):
+            if not self.check_compatibility_flag("category_search_yields_nothing"):
                 if self.check_compatibility_flag("text_search_is_case_insensitive"):
                     assert len(some_events) == 1
                 else:
@@ -1184,51 +1278,52 @@ class RepeatedFunctionalTestsBaseClass(object):
             some_events = c.search(
                 comp_class=Event, category="ANNIVERSARY,PERSONAL,SPECIAL OCCASION"
             )
-            if not self.check_compatibility_flag("text_search_not_working"):
-                assert len(some_events) in (0, 1)
+            assert len(some_events) in (0, 1)
             ## TODO: This is actually a bug. We need to do client side filtering
             some_events = c.search(comp_class=Event, category="PERSON")
+            if self.check_compatibility_flag("text_search_is_exact_match_sometimes"):
+                assert len(some_events) in (0, 1)
             if self.check_compatibility_flag("text_search_is_exact_match_only"):
                 assert len(some_events) == 0
-            elif not self.check_compatibility_flag(
-                "category_search_yields_nothing"
-            ) and not self.check_compatibility_flag("text_search_not_working"):
+            elif not self.check_compatibility_flag("category_search_yields_nothing"):
                 assert len(some_events) == 1
 
-            ## I expect logical and when combining category with a date range
+            ## I expect "logical and" when combining category with a date range
             no_events = c.search(
                 comp_class=Event,
                 category="PERSONAL",
                 start=datetime(2006, 7, 13, 13, 0),
                 end=datetime(2006, 7, 15, 13, 0),
             )
-            if (
-                not self.check_compatibility_flag("category_search_yields_nothing")
-                and not self.check_compatibility_flag("combined_search_not_working")
-                and not self.check_compatibility_flag("text_search_not_working")
-            ):
-                assert len(no_events) == 0
+            if not self.check_compatibility_flag(
+                "category_search_yields_nothing"
+            ) and not self.check_compatibility_flag("combined_search_not_working"):
+                if self.check_compatibility_flag("fastmail_buggy_noexpand_date_search"):
+                    ## fastmail and davical delivers too many recurring events on a date search
+                    ## (but fastmail anyway won't get here, as combined search is not working with fastmail)
+                    assert len(no_events) == 1
+                else:
+                    assert len(no_events) == 1
             some_events = c.search(
                 comp_class=Event,
                 category="PERSONAL",
                 start=datetime(1997, 11, 1, 13, 0),
                 end=datetime(1997, 11, 3, 13, 0),
             )
-            if (
-                not self.check_compatibility_flag("category_search_yields_nothing")
-                and not self.check_compatibility_flag("combined_search_not_working")
-                and not self.check_compatibility_flag("text_search_not_working")
-            ):
+            if not self.check_compatibility_flag(
+                "category_search_yields_nothing"
+            ) and not self.check_compatibility_flag("combined_search_not_working"):
                 assert len(some_events) == 1
 
         some_events = c.search(comp_class=Event, summary="Bastille Day Party")
-        if not self.check_compatibility_flag("text_search_not_working"):
-            if self.check_compatibility_flag("text_search_is_exact_match_sometimes"):
-                assert len(some_events) in (1, 2)
-            elif self.check_compatibility_flag("text_search_is_exact_match_only"):
-                assert len(some_events) == 1
-            else:
-                assert len(some_events) == 2
+        assert len(some_events) == 1
+        some_events = c.search(comp_class=Event, summary="Bastille Day")
+        if self.check_compatibility_flag("text_search_is_exact_match_sometimes"):
+            assert len(some_events) in (0, 2)
+        elif self.check_compatibility_flag("text_search_is_exact_match_only"):
+            assert len(some_events) == 0
+        else:
+            assert len(some_events) == 2
 
         ## Even sorting should work out
         all_events = c.search(sort_keys=("summary", "dtstamp"))
@@ -1249,15 +1344,24 @@ class RepeatedFunctionalTestsBaseClass(object):
 
         ## Search without any parameters should yield everything on calendar
         all_todos = c.search()
-        assert len(all_todos) == 6
+        if self.check_compatibility_flag("search_needs_comptype"):
+            assert len(all_todos) <= 6
+        else:
+            assert len(all_todos) == 6
 
         ## Search with comp_class set to Event should yield all events on calendar
         all_todos = c.search(comp_class=Event)
         assert len(all_todos) == 0
 
-        ## Search with todo flag set should yield all 6 events
+        ## Search with todo flag set should yield all 6 tasks
+        ## (Except, if the calendar server does not support is-not-defined very
+        ## well, perhaps only 3 will be returned - see
+        ## https://gitlab.com/davical-project/davical/-/issues/281 )
         all_todos = c.search(todo=True)
-        assert len(all_todos) == 6
+        if self.check_compatibility_flag("isnotdefined_not_working"):
+            assert len(all_todos) in (3, 6)
+        else:
+            assert len(all_todos) == 6
 
         ## Search for misc text fields
         ## UID is a special case, supported by almost all servers
@@ -1291,10 +1395,11 @@ class RepeatedFunctionalTestsBaseClass(object):
         ## This is not a very useful search, and it's sort of a client side bug that we allow it at all.
         ## It will not match if categories field is set to "PERSONAL,ANNIVERSARY,SPECIAL OCCATION"
         ## It may not match since the above is to be considered equivalent to the raw data entered.
-        some_todos = c.search(comp_class=Event, category="FAMILY,FINANCE")
+        some_todos = c.search(comp_class=Todo, category="FAMILY,FINANCE")
         if not self.check_compatibility_flag("text_search_not_working"):
-            assert len(some_todos) in (0, 1)
-        ## TODO: This is actually a bug. We need to do client side filtering
+            assert len(some_todos) in (0, 6)
+        ## TODO: We should consider to do client side filtering to ensure exact
+        ## match only on components having MIL as a category (and not FAMILY)
         some_todos = c.search(comp_class=Todo, category="MIL")
         if self.check_compatibility_flag("text_search_is_exact_match_sometimes"):
             assert len(some_todos) in (0, 6)
@@ -1303,16 +1408,16 @@ class RepeatedFunctionalTestsBaseClass(object):
         elif not self.check_compatibility_flag(
             "category_search_yields_nothing"
         ) and not self.check_compatibility_flag("text_search_not_working"):
-            assert len(some_todos) == 1
+            ## This is the correct thing, according to the letter of the RFC
+            assert len(some_todos) == 6
 
-        ## completing an event, and it should not show up anymore
-        t1.complete()
-        t2.complete()
+        ## completing events, and it should not show up anymore
         t3.complete()
-        t4.complete()
+        t5.complete()
+        t6.complete()
 
         some_todos = c.search(todo=True)
-        assert len(some_todos) == 2
+        assert len(some_todos) == 3
 
         ## unless we specifically ask for completed tasks
         all_todos = c.search(todo=True, include_completed=True)
@@ -1326,18 +1431,21 @@ class RepeatedFunctionalTestsBaseClass(object):
             dtstart=datetime(2022, 12, 26, 19, 15),
             dtend=datetime(2022, 12, 26, 20, 00),
             summary="this is a parent event test",
+            uid="ctuid1",
         )
         child = c.save_event(
             dtstart=datetime(2022, 12, 26, 19, 17),
             dtend=datetime(2022, 12, 26, 20, 00),
             summary="this is a child event test",
             parent=[parent.id],
+            uid="ctuid2",
         )
         grandparent = c.save_event(
             dtstart=datetime(2022, 12, 26, 19, 00),
             dtend=datetime(2022, 12, 26, 20, 00),
             summary="this is a grandparent event test",
             child=[parent.id],
+            uid="ctuid3",
         )
 
         parent_ = c.event_by_uid(parent.id)
@@ -1363,6 +1471,113 @@ class RepeatedFunctionalTestsBaseClass(object):
         assert rt == parent.id
         assert rt.params["RELTYPE"] == "PARENT"
 
+    def testSetDue(self):
+        self.skip_on_compatibility_flag("read_only")
+
+        c = self._fixCalendar(supported_calendar_component_set=["VTODO"])
+
+        utc = timezone.utc
+
+        some_todo = c.save_todo(
+            dtstart=datetime(2022, 12, 26, 19, 15, tzinfo=utc),
+            due=datetime(2022, 12, 26, 20, 00, tzinfo=utc),
+            summary="Some task",
+            uid="ctuid1",
+        )
+
+        ## setting the due should ... set the due (surprise, surprise)
+        some_todo.set_due(datetime(2022, 12, 26, 20, 10, tzinfo=utc))
+        assert some_todo.icalendar_component["DUE"].dt == datetime(
+            2022, 12, 26, 20, 10, tzinfo=utc
+        )
+        assert some_todo.icalendar_component["DTSTART"].dt == datetime(
+            2022, 12, 26, 19, 15, tzinfo=utc
+        )
+
+        ## move_dtstart causes the duration to be unchanged
+        some_todo.set_due(datetime(2022, 12, 26, 20, 20, tzinfo=utc), move_dtstart=True)
+        assert some_todo.icalendar_component["DUE"].dt == datetime(
+            2022, 12, 26, 20, 20, tzinfo=utc
+        )
+        assert some_todo.icalendar_component["DTSTART"].dt == datetime(
+            2022, 12, 26, 19, 25, tzinfo=utc
+        )
+
+        ## This task has duration set rather than due.  Due should be implied to be 19:30.
+        some_other_todo = c.save_todo(
+            dtstart=datetime(2022, 12, 26, 19, 15, tzinfo=utc),
+            duration=timedelta(minutes=15),
+            summary="Some other task",
+            uid="ctuid2",
+        )
+        some_other_todo.set_due(
+            datetime(2022, 12, 26, 19, 45, tzinfo=utc), move_dtstart=True
+        )
+        assert some_other_todo.icalendar_component["DUE"].dt == datetime(
+            2022, 12, 26, 19, 45, tzinfo=utc
+        )
+        assert some_other_todo.icalendar_component["DTSTART"].dt == datetime(
+            2022, 12, 26, 19, 30, tzinfo=utc
+        )
+
+        some_todo.save()
+
+        self.skip_on_compatibility_flag("no_relships")
+        parent = c.save_todo(
+            dtstart=datetime(2022, 12, 26, 19, 00, tzinfo=utc),
+            dtend=datetime(2022, 12, 26, 21, 00, tzinfo=utc),
+            summary="this is a parent test task",
+            uid="ctuid3",
+            child=[some_todo.id],
+        )
+
+        ## The above updates the some_todo object on the server side, but the local object is not
+        ## updated ... until we reload it
+        some_todo.load()
+
+        ## This should work out (set the childs due to some time before the parents due)
+        some_todo.set_due(
+            datetime(2022, 12, 26, 20, 30, tzinfo=utc),
+            move_dtstart=True,
+            check_dependent=True,
+        )
+        assert some_todo.icalendar_component["DUE"].dt == datetime(
+            2022, 12, 26, 20, 30, tzinfo=utc
+        )
+        assert some_todo.icalendar_component["DTSTART"].dt == datetime(
+            2022, 12, 26, 19, 35, tzinfo=utc
+        )
+
+        ## This should not work out (set the childs due to some time before the parents due)
+        with pytest.raises(error.ConsistencyError):
+            some_todo.set_due(
+                datetime(2022, 12, 26, 21, 30, tzinfo=utc),
+                move_dtstart=True,
+                check_dependent=True,
+            )
+
+        child = c.save_todo(
+            dtstart=datetime(2022, 12, 26, 19, 45),
+            due=datetime(2022, 12, 26, 19, 55),
+            summary="this is a test child task",
+            uid="ctuid4",
+            parent=[some_todo.id],
+        )
+
+        ## This should still work out (set the childs due to some time before the parents due)
+        ## (The fact that we now have a child does not affect it anyhow)
+        some_todo.set_due(
+            datetime(2022, 12, 26, 20, 31, tzinfo=utc),
+            move_dtstart=True,
+            check_dependent=True,
+        )
+        assert some_todo.icalendar_component["DUE"].dt == datetime(
+            2022, 12, 26, 20, 31, tzinfo=utc
+        )
+        assert some_todo.icalendar_component["DTSTART"].dt == datetime(
+            2022, 12, 26, 19, 36, tzinfo=utc
+        )
+
     def testCreateJournalListAndJournalEntry(self):
         """
         This test demonstrates the support for journals.
@@ -1384,6 +1599,7 @@ class RepeatedFunctionalTestsBaseClass(object):
             dtstart=date(2011, 11, 11),
             summary="A childbirth in a hospital in Kupchino",
             description="A quick birth, in the middle of the night",
+            uid="ctuid1",
         )
         assert len(c.journals()) == 2
         todos = c.todos()
@@ -1425,17 +1641,20 @@ class RepeatedFunctionalTestsBaseClass(object):
         assert len(todos) == 1
         assert len(todos2) == 1
 
-        t3 = c.save_todo(summary="mop the floor", categories=["housework"], priority=4)
+        t3 = c.save_todo(
+            summary="mop the floor", categories=["housework"], priority=4, uid="ctuid1"
+        )
         assert len(c.todos()) == 2
 
         # adding a todo without an UID, it should also work (library will add the missing UID)
-        c.save_todo(todo7)
+        t7 = c.save_todo(todo7)
         assert len(c.todos()) == 3
 
         logging.info("Fetching the events (should be none)")
         # c.events() should NOT return todo-items
         events = c.events()
         assert len(events) == 0
+        t7.delete()
 
     def testTodos(self):
         """
@@ -1452,7 +1671,7 @@ class RepeatedFunctionalTestsBaseClass(object):
         # add todo-item
         t1 = c.save_todo(todo)
         t2 = c.save_todo(todo2)
-        t3 = c.save_todo(todo3)
+        t4 = c.save_todo(todo4)
 
         todos = c.todos()
         assert len(todos) == 3
@@ -1460,7 +1679,8 @@ class RepeatedFunctionalTestsBaseClass(object):
         def uids(lst):
             return [x.instance.vtodo.uid for x in lst]
 
-        assert uids(todos) == uids([t2, t3, t1])
+        ## Default sort order is (due, priority).
+        assert uids(todos) == uids([t2, t1, t4])
 
         todos = c.todos(sort_keys=("priority",))
         ## sort_key is considered to be a legacy parameter,
@@ -1474,8 +1694,8 @@ class RepeatedFunctionalTestsBaseClass(object):
                 if hasattr(x.instance.vtodo, "priority")
             ]
 
-        assert pri(todos) == pri([t3, t2])
-        assert pri(todos2) == pri([t3, t2])
+        assert pri(todos) == pri([t4, t2])
+        assert pri(todos2) == pri([t4, t2])
 
         todos = c.todos(
             sort_keys=(
@@ -1483,7 +1703,7 @@ class RepeatedFunctionalTestsBaseClass(object):
                 "priority",
             )
         )
-        assert uids(todos) == uids([t3, t2, t1])
+        assert uids(todos) == uids([t4, t2, t1])
 
         ## str of CalendarObjectResource is slightly inconsistent compared to
         ## the str of Calendar objects, as the class name is included.  Perhaps
@@ -1511,7 +1731,10 @@ class RepeatedFunctionalTestsBaseClass(object):
         t5 = c.save_todo(todo5)
         t6 = c.save_todo(todo6)
         todos = c.todos()
-        assert len(todos) == 6
+        if self.check_compatibility_flag("isnotdefined_not_working"):
+            assert len(todos) in (3, 6)
+        else:
+            assert len(todos) == 6
 
         notodos = c.date_search(  # default compfilter is events
             start=datetime(1997, 4, 14), end=datetime(2015, 5, 14), expand=False
@@ -1632,7 +1855,7 @@ class RepeatedFunctionalTestsBaseClass(object):
         # add todo-items
         t1 = c.save_todo(todo)
         t2 = c.save_todo(todo2)
-        t3 = c.save_todo(todo3)
+        t3 = c.save_todo(todo3, status="NEEDS-ACTION")
 
         # There are now three todo-items at the calendar
         todos = c.todos()
@@ -1670,7 +1893,7 @@ class RepeatedFunctionalTestsBaseClass(object):
     def testTodoRecurringCompleteSafe(self):
         self.skip_on_compatibility_flag("read_only")
         c = self._fixCalendar(supported_calendar_component_set=["VTODO"])
-        t6 = c.save_todo(todo6)
+        t6 = c.save_todo(todo6, status="NEEDS-ACTION")
         if not self.check_compatibility_flag("rrule_takes_no_count"):
             t8 = c.save_todo(todo8)
         if not self.check_compatibility_flag("rrule_takes_no_count"):
@@ -1681,19 +1904,23 @@ class RepeatedFunctionalTestsBaseClass(object):
         if self.check_compatibility_flag("rrule_takes_no_count"):
             assert len(c.todos()) == 1
             assert len(c.todos(include_completed=True)) == 2
+            c.todos()[0].delete()
         self.skip_on_compatibility_flag("rrule_takes_no_count")
         assert len(c.todos()) == 2
         assert len(c.todos(include_completed=True)) == 3
         t8.complete(handle_rrule=True, rrule_mode="safe")
-        assert len(c.todos()) == 2
+        todos = c.todos()
+        assert len(todos) == 2
         t8.complete(handle_rrule=True, rrule_mode="safe")
         t8.complete(handle_rrule=True, rrule_mode="safe")
         assert len(c.todos()) == 1
+        assert len(c.todos(include_completed=True)) == 5
+        [x.delete() for x in c.todos(include_completed=True)]
 
     def testTodoRecurringCompleteThisandfuture(self):
         self.skip_on_compatibility_flag("read_only")
         c = self._fixCalendar(supported_calendar_component_set=["VTODO"])
-        t6 = c.save_todo(todo6)
+        t6 = c.save_todo(todo6, status="NEEDS-ACTION")
         if not self.check_compatibility_flag("rrule_takes_no_count"):
             t8 = c.save_todo(todo8)
         if not self.check_compatibility_flag("rrule_takes_no_count"):
@@ -1721,6 +1948,11 @@ class RepeatedFunctionalTestsBaseClass(object):
         # TODO: split up in creating a calendar with non-ascii name
         # and an event with non-ascii description
         self.skip_on_compatibility_flag("no_mkcalendar")
+        if not self.check_compatibility_flag(
+            "unique_calendar_ids"
+        ) and self.cleanup_regime in ("light", "pre"):
+            self._teardownCalendar(cal_id=self.testcal_id)
+
         c = self.principal.make_calendar(name="Yølp", cal_id=self.testcal_id)
 
         # add event
@@ -1740,9 +1972,19 @@ class RepeatedFunctionalTestsBaseClass(object):
         if "zimbra" not in str(c.url):
             assert len(events) == 1
 
+        if (
+            not self.check_compatibility_flag("unique_calendar_ids")
+            and self.cleanup_regime == "post"
+        ):
+            self._teardownCalendar(cal_id=self.testcal_id)
+
     def testUnicodeEvent(self):
         self.skip_on_compatibility_flag("read_only")
         self.skip_on_compatibility_flag("no_mkcalendar")
+        if not self.check_compatibility_flag(
+            "unique_calendar_ids"
+        ) and self.cleanup_regime in ("light", "pre"):
+            self._teardownCalendar(cal_id=self.testcal_id)
         c = self.principal.make_calendar(name="Yølp", cal_id=self.testcal_id)
 
         # add event
@@ -1777,6 +2019,10 @@ class RepeatedFunctionalTestsBaseClass(object):
 
         # Creating a new calendar with different ID but with existing name
         # TODO: why do we do this?
+        if not self.check_compatibility_flag(
+            "unique_calendar_ids"
+        ) and self.cleanup_regime in ("light", "pre"):
+            self._teardownCalendar(cal_id=self.testcal_id2)
         cc = self.principal.make_calendar("Yep", self.testcal_id2)
         cc.delete()
 
@@ -2137,91 +2383,6 @@ class RepeatedFunctionalTestsBaseClass(object):
             principal = conn.principal()
             calendars = principal.calendars()
 
-    ## TODO: run this test, ref https://github.com/python-caldav/caldav/issues/91
-    ## It should be removed prior to a 1.0-release.
-    def testBackwardCompatibility(self):
-        """
-        Tobias Brox has done some API changes - but this thing should
-        still be backward compatible.
-        """
-        self.skip_on_compatibility_flag("read_only")
-        if "backwards_compatibility_url" not in self.server_params:
-            pytest.skip(
-                "backward compatibility check skipped - needs an URL like it was supposed to be in 2013"
-            )
-        caldav = DAVClient(self.server_params["backwards_compatibility_url"])
-        principal = Principal(caldav, self.server_params["backwards_compatibility_url"])
-        c = Calendar(caldav, name="Yep", parent=principal, id=self.testcal_id).save()
-        assert c.url is not None
-
-        c.set_properties(
-            [
-                dav.DisplayName("hooray"),
-            ]
-        )
-        props = c.get_properties(
-            [
-                dav.DisplayName(),
-            ]
-        )
-        assert props[dav.DisplayName.tag] == "hooray"
-
-        cc = Calendar(caldav, name="Yep", parent=principal).save()
-        assert cc.url is not None
-        cc.delete()
-
-        e = Event(caldav, data=ev1, parent=c).save()
-        assert e.url is not None
-        ee = Event(caldav, url=url.make(e.url), parent=c)
-        ee.load()
-        assert e.instance.vevent.uid == ee.instance.vevent.uid
-
-        r = c.date_search(
-            datetime(2006, 7, 13, 17, 00, 00),
-            datetime(2006, 7, 15, 17, 00, 00),
-            expand=False,
-        )
-        assert e.instance.vevent.uid == r[0].instance.vevent.uid
-        assert len(r) == 1
-
-        all = c.events()
-        assert len(all) == 1
-
-        e2 = Event(caldav, data=ev2, parent=c).save()
-        assert e.url is not None
-
-        tmp = c.event("20010712T182145Z-123401@example.com")
-        assert e2.instance.vevent.uid == tmp.instance.vevent.uid
-
-        r = c.date_search(
-            datetime(2007, 7, 13, 17, 00, 00),
-            datetime(2007, 7, 15, 17, 00, 00),
-            expand=False,
-        )
-        assert len(r) == 1
-
-        e.data = ev2
-        e.save()
-
-        r = c.date_search(
-            datetime(2007, 7, 13, 17, 00, 00),
-            datetime(2007, 7, 15, 17, 00, 00),
-            expand=False,
-        )
-        # for e in r: print(e.data)
-        assert len(r) == 1
-
-        e.instance = e2.instance
-        e.save()
-
-        r = c.date_search(
-            datetime(2007, 7, 13, 17, 00, 00),
-            datetime(2007, 7, 15, 17, 00, 00),
-            expand=False,
-        )
-        # for e in r: print(e.data)
-        assert len(r) == 1
-
     def testObjects(self):
         # TODO: description ... what are we trying to test for here?
         o = DAVObject(self.caldav)
